
using System;
using System.Collections.Generic;
using RealTime.CustomAI;
using RealTime.GameConnection;
using UnityEngine;

namespace RealTime.Pandemic
{
    public class PandemicManager : MonoBehaviour
    {
        private bool QUICK_SIMULATION = true;
        private bool ACTIVE_ONLY = true;

        private System.Random random = new System.Random(0);
        private bool active = false;
        private float spreadDistance = 10;
        private DateTime lastDateTime;
        private DateTime lastDateTimeCitizensUpdate;
        private DateTime lastStoreTime;
        private DateTime currentDateTime;

        private List<uint> initialPopulation = new List<uint>();
        private List<uint> initialPopulationHealthy = new List<uint>();
        private List<uint> initialPopulationRecovered = new List<uint>();
        private List<uint> initialPopulationDead = new List<uint>();
        private List<uint> initialPopulationSick = new List<uint>();

        private List<uint> removedCitizens = new List<uint>();
        private HashSet<uint> usedCitizens = new HashSet<uint>();
        private Dictionary<uint, uint> citizenMatching = new Dictionary<uint, uint>();

        private PandemicObserver Observer;

        private Dictionary<uint, long> activeInfections = new Dictionary<uint, long>();
        private HashSet<uint> infectedCitizens = new HashSet<uint>();
        private HashSet<uint> infectedCitizensWithSymptoms = new HashSet<uint>();

        private SimulationManager simulation;

        private readonly long CITIZENS_UPDATE_INTERVAL_MINUTES = 60 * 6;
        private readonly long UPDATE_INTERVAL_MINUTES = 5;
        private readonly float STORE_INTERVAL_DAYS = 1;

        private MaskManager Masks = new MaskManager();

        public PandemicManager()
        {
            Instance = this;
        }

        internal void Init(Config.RealTimeConfig config, GameConnections<Citizen> connections)
        {
            Observer = new PandemicObserver();

            CitizenMgr = connections.CitizenManager;
            CitizenProxy = connections.CitizenConnection;
            BuildingMgr = connections.BuildingManager;

            Config = config;

            active = true;

            ContactManager.Instance.Init(config);
        }

        public long GetInfectionDate(uint citizenID)
        {
            if (!activeInfections.ContainsKey(citizenID))
            {
                return -1;
            }
            return activeInfections[citizenID];
        }

        private void CloseSchools()
        {
            for (ushort i = 0; i < BuildingMgr.GetMaxBuildingsCount(); i++)
            {
                if (BuildingMgr.GetBuildingService(i) == ItemClass.Service.Education)
                {
                    BuildingMgr.ReleaseBuilding(i);
                }
            }
        }

        public void Start()
        {
            simulation = GameObject.Find("SimulationManager").GetComponent<SimulationManager>();
            currentDateTime = simulation.m_currentGameTime;

            Citizen[] citizens = CitizenMgr.GetCitizensArray();

            var infectionCandidates = new List<uint>();

            for (uint i = 0; i < citizens.Length; i++)
            {
                if (CitizenProxy.IsEmpty(ref citizens[retrieveID(i)]) || CitizenProxy.GetHomeBuilding(ref citizens[retrieveID(i)]) == 0)
                {
                    continue;
                }

                initialPopulation.Add(i);
                usedCitizens.Add(i);

                if (CitizenProxy.IsSick(ref citizens[retrieveID(i)]))
                {
                    CitizenProxy.SetSick(ref citizens[retrieveID(i)], false);
                }

                initialPopulationHealthy.Add(i);
                if (CitizenProxy.GetLocation(ref citizens[retrieveID(i)]) == Citizen.Location.Home)
                {
                    infectionCandidates.Add(i);
                }
            }

            float infectionRatio = Config.DiseaseStartInfectionRatio;
            uint intendedNumberOfSickCitizens = Math.Max((uint)(infectionRatio * (initialPopulationSick.Count + initialPopulationHealthy.Count) / 100f), 1);

            Debug.Log("Intended number of sick citizens: " + intendedNumberOfSickCitizens + ", actual number of sick citizens: " + initialPopulationSick.Count + ", number of healthy citizens: " + initialPopulationHealthy.Count + ", infection candidates: " + infectionCandidates);

            while (intendedNumberOfSickCitizens > initialPopulationSick.Count)
            {
                // Sickening citizens
                uint citizenID;
                if (infectionCandidates.Count > 0)
                {
                    int index = random.Next(infectionCandidates.Count);
                    citizenID = infectionCandidates[index];
                    infectionCandidates.RemoveAt(index);
                    initialPopulationHealthy.Remove(citizenID);
                } else
                {
                    int index = random.Next(initialPopulationHealthy.Count);
                    citizenID = initialPopulationHealthy[index];
                    initialPopulationHealthy.RemoveAt(index);
                }
                int offset = random.Next((int) Math.Max(Config.StartInfection - 1, 0) * 24 * 3600 * 1000);
                InfectCitizen(citizenID, ref citizens[retrieveID(citizenID)], -offset);
            }
            Debug.Log("Intended number of sick citizens: " + intendedNumberOfSickCitizens + ", actual number of sick citizens: " + initialPopulationSick.Count + ", number of healthy citizens: " + initialPopulationHealthy.Count);

            lastDateTime = simulation.m_currentGameTime;
            lastDateTimeCitizensUpdate = simulation.m_currentGameTime;
            lastStoreTime = simulation.m_currentGameTime;

            QuarantineManager.Instance.InLockDown = true;

            TestManager.Instance.Init(Config, initialPopulation.Count, currentDateTime);
            Masks.Init(Config);
        }

        public void Update()
        {
            DateTime tempDateTime = simulation.m_currentGameTime;
            if ((tempDateTime.Ticks - lastDateTime.Ticks) / TimeSpan.TicksPerMinute < UPDATE_INTERVAL_MINUTES) {
                return;
            }

            currentDateTime = tempDateTime;
            simulation.ForcedSimulationPaused = true;

            long milliseconds = (currentDateTime.Ticks - lastDateTime.Ticks) / TimeSpan.TicksPerMillisecond;

            Masks.SetStepLengthInHours(milliseconds / (3600 * 1000.0));

            double childDeathProbability = 1 - Math.Pow(1 - Config.DeathChild / 100.0, milliseconds / ((Config.DiseaseDuration - Config.StartSymptoms) * 24 * 3600 * 1000.0));
            double teenDeathProbability = 1 - Math.Pow(1 - Config.DeathTeen / 100.0, milliseconds / ((Config.DiseaseDuration - Config.StartSymptoms) * 24 * 3600 * 1000.0));
            double youngDeathProbability = 1 - Math.Pow(1 - Config.DeathYoung / 100.0, milliseconds / ((Config.DiseaseDuration - Config.StartSymptoms) * 24 * 3600 * 1000.0));
            double adultDeathProbability = 1 - Math.Pow(1 - Config.DeathAdult / 100.0, milliseconds / ((Config.DiseaseDuration - Config.StartSymptoms) * 24 * 3600 * 1000.0));
            double seniorDeathProbability = 1 - Math.Pow(1 - Config.DeathSenior / 100.0, milliseconds / ((Config.DiseaseDuration - Config.StartSymptoms) * 24 * 3600 * 1000.0));

            kill(childDeathProbability, teenDeathProbability, youngDeathProbability, adultDeathProbability, seniorDeathProbability);

            float modificationOutdoor = 1;
            float modificationIndoor = 1;
            float modificationVehicle = 1;
            if (Config.MaskBehavior == RealTime.Config.MaskBehavior.Full)
            {
                modificationOutdoor = Config.TransmissionProbabilityReduction;
                modificationIndoor = Config.TransmissionProbabilityReduction;
                modificationVehicle = Config.TransmissionProbabilityReduction;
            }
            else if (Config.MaskBehavior == RealTime.Config.MaskBehavior.Vehicle)
            {
                modificationIndoor = Config.TransmissionProbabilityReduction;
                modificationVehicle = Config.TransmissionProbabilityReduction;
            }
            else if (Config.MaskBehavior == RealTime.Config.MaskBehavior.Building)
            {
                modificationIndoor = Config.TransmissionProbabilityReduction;
            }

            double outdoorTransmissionProbability = 1 - Math.Pow(1 - (Config.OutdoorDiseaseTransmissionProbability / 100.0 / modificationOutdoor), milliseconds / (3600 * 1000.0));
            double indoorTransmissionProbability = 1 - Math.Pow(1 - (Config.IndoorDiseaseTransmissionProbability / 100.0 / modificationIndoor), milliseconds / (3600 * 1000.0));
            double indoorInfectionProbabilityNoContact = 1 - Math.Pow(1 - (Config.IndoorDiseaseTransmissionProbability / 100.0 / 96.0), milliseconds / (3600 * 1000.0));
            double indoorDefaultTransmissionProbability = 1 - Math.Pow(1 - (Config.IndoorDiseaseTransmissionProbability / 100.0), milliseconds / (3600 * 1000.0));
            double vehicleTransmissionProbability = 1 - Math.Pow(1 - (Config.IndoorDiseaseTransmissionProbability / 100.0 / modificationVehicle), milliseconds / (3600 * 1000.0));

            if (ACTIVE_ONLY)
            {
                if ((tempDateTime.Ticks - lastDateTimeCitizensUpdate.Ticks) / TimeSpan.TicksPerMinute >= CITIZENS_UPDATE_INTERVAL_MINUTES)
                {
                    Debug.Log("Number of inactive citizens before: " + removedCitizens.Count);

                    lastDateTimeCitizensUpdate = tempDateTime;
                    Citizen[] citizens = CitizenMgr.GetCitizensArray();

                    for (int i = 0; i < initialPopulation.Count; i++)
                    {
                        if (CitizenProxy.IsEmpty(ref citizens[retrieveID(initialPopulation[i])]))
                        {
                            if (!initialPopulationDead.Contains(initialPopulation[i]) && !initialPopulationRecovered.Contains(initialPopulation[i]))
                            {
                                if (!removedCitizens.Contains(initialPopulation[i]))
                                {
                                    removedCitizens.Add(initialPopulation[i]);
                                }
                            }
                            usedCitizens.Remove(retrieveID(initialPopulation[i]));
                            citizenMatching.Remove(initialPopulation[i]);
                            initialPopulation.RemoveAt(i);

                            i--;

                        }
                    }
                    for (uint i = 0; i < citizens.Length && removedCitizens.Count > 0; i++)
                    {
                        if (!CitizenProxy.IsEmpty(ref citizens[i]) && CitizenProxy.GetHomeBuilding(ref citizens[i]) != 0 && !usedCitizens.Contains(i))
                        {
                            if (removedCitizens.Count > 0)
                            {
                                uint removedCitizen = removedCitizens[0];
                                removedCitizens.RemoveAt(0);

                                usedCitizens.Add(i);
                                citizenMatching.Add(removedCitizen, i);
                                initialPopulation.Add(removedCitizen);

                                Debug.Log("Replaced " + removedCitizen + " with citizen " + i);
                            }
                        }
                    }
                }
                Debug.Log("Number of inactive citizens: " + removedCitizens.Count);
            }

            spread();

            Observer.AddSickCitizens(currentDateTime, initialPopulationSick.Count);
            Observer.AddHealthyCitizens(currentDateTime, initialPopulationHealthy.Count);
            Observer.AddRecoveredCitizens(currentDateTime, initialPopulationRecovered.Count);
            Observer.AddDeadCitizens(currentDateTime, initialPopulationDead.Count);

            Observer.FinishObservations(currentDateTime);

            Debug.Log("Sick: " + Observer.GetSickCitizens() + ", Healty: " + Observer.GetHealthyCitizens() + ", Recovered: " + Observer.GetRecoveredCitizens() + ", Dead: " + Observer.GetDeadCitizens());
            Debug.Log("Citizens in quarantine: " + QuarantineManager.Instance.CitizensInQuarantine());

            lastDateTime = currentDateTime;

            if ((tempDateTime.Ticks - lastStoreTime.Ticks) / TimeSpan.TicksPerDay < STORE_INTERVAL_DAYS)
            {
                try
                {
                    Observer.WriteToDisc(true);
                }
                catch (Exception e)
                {

                }
                lastStoreTime = tempDateTime;
            }

            if (Observer.GetSickCitizens() == 0)
            {
                Debug.Log("Writing to disc");

                Observer.WriteToDisc(false);

                simulation.SimulationPaused = true;
            }

            simulation.ForcedSimulationPaused = false;
        }

        private bool IsInfectious(uint citizenID)
        {
            if (activeInfections.ContainsKey(citizenID))
            {
                long infectedTime = (currentDateTime.Ticks / 10000) - activeInfections[citizenID];
                double infectedTimeInDays = infectedTime / 1000.0 / 3600.0 / 24.0;
                if (Config.StartInfection <= infectedTimeInDays && infectedTimeInDays <= Config.EndInfection)
                {
                    return true;
                }
            }
            return false;
        }

        private void InfectCitizen(uint infectedCitizenID, ref Citizen infectedCitizen)
        {
            InfectCitizen(infectedCitizenID, ref infectedCitizen, 0);
        }

        private void InfectCitizen(uint infectedCitizenID, ref Citizen infectedCitizen, long offset)
        {
            if (infectedCitizens.Contains(infectedCitizenID))
            {
                return;
            }

            CitizenProxy.SetSick(ref infectedCitizen, true);
            infectedCitizens.Add(infectedCitizenID);

            if (random.NextDouble() < Config.SymptomProbability / 100.0)
            {
                infectedCitizensWithSymptoms.Add(infectedCitizenID);
            }

            long timeOfInfection = currentDateTime.Ticks / 10000;
            activeInfections.Add(infectedCitizenID, timeOfInfection + offset);

            initialPopulationHealthy.Remove(infectedCitizenID);
            initialPopulationSick.Add(infectedCitizenID);
        }

        private void HealCitizen(uint infectedCitizenID, ref Citizen infectedCitizen)
        {
            if (!infectedCitizens.Contains(infectedCitizenID))
            {
                return;
            }

            CitizenProxy.SetSick(ref infectedCitizen, false);
            activeInfections.Remove(infectedCitizenID);

            initialPopulationRecovered.Add(infectedCitizenID);
            initialPopulationSick.Remove(infectedCitizenID);
        }

        private void KillCitizen(uint infectedCitizenID, ref Citizen infectedCitizen)
        {
            if (!infectedCitizens.Contains(infectedCitizenID) || !activeInfections.ContainsKey(infectedCitizenID))
            {
                return;
            }

            CitizenProxy.SetDead(ref infectedCitizen, true);
            activeInfections.Remove(infectedCitizenID);

            initialPopulationDead.Add(infectedCitizenID);
            initialPopulationSick.Remove(infectedCitizenID);
        }

        private bool IsInQuarantine(uint citizenId)
        {
            return QuarantineManager.Instance.IsInQuarantine(citizenId, currentDateTime) || QuarantineManager.Instance.IsInProphylacticQuarantine(citizenId, currentDateTime);
        }

        public void Shuffle<T>(IList<T> list)
        {
            int n = list.Count;
            while (n > 1)
            {
                n--;
                int k = random.Next(n + 1);
                T value = list[k];
                list[k] = list[n];
                list[n] = value;
            }
        }

        public void spread()
        {
            Debug.Log("Spreading disease");

            Citizen[] citizens = CitizenMgr.GetCitizensArray();

            HashSet<uint> newlyInfectedCitizenIDs = new HashSet<uint>();
            
            float range = Config.DiseaseTransmissionRange;

            uint numberOfSickCitizens = 0;
            uint numberOfInfectiousCitizens = 0;

            Shuffle(initialPopulation);
            for (int i = 0; i < initialPopulation.Count; i++)
            {
                uint citizenID = initialPopulation[i];

                if (!CitizenProxy.IsDead(ref citizens[retrieveID(citizenID)]))
                {
                    TestManager.Instance.TestCitizen(citizenID, CitizenProxy.IsSick(ref citizens[retrieveID(citizenID)]), IsKnownSick(citizenID), currentDateTime);
                }
            }

            uint tested = 0;
            for (int i = 0; i < initialPopulationSick.Count; i++)
            {
                uint citizenID = initialPopulationSick[i];

                if (newlyInfectedCitizenIDs.Contains(citizenID))
                {
                    continue;
                }

                ushort instanceID = CitizenProxy.GetInstance(ref citizens[retrieveID(citizenID)]);

                if (!CitizenProxy.IsSick(ref citizens[retrieveID(citizenID)]) && activeInfections.ContainsKey(citizenID))
                {
                    CitizenProxy.SetSick(ref citizens[retrieveID(citizenID)], true);
                }

                double infectedTimeInDays = 0;
                if (activeInfections.ContainsKey(citizenID))
                {
                    long infectedTime = (currentDateTime.Ticks / 10000) - activeInfections[citizenID];
                    infectedTimeInDays = infectedTime / 1000.0 / 3600.0 / 24.0;
                    if (infectedTimeInDays > Config.DiseaseDuration)
                    {
                        HealCitizen(citizenID, ref citizens[retrieveID(citizenID)]);
                        i--;
                        continue;
                    }
                }

                if (ShouldBeInQuarantine(citizenID) && (Config.QuarantineBehavior == RealTime.Config.QuarantineBehavior.Contacts || Config.QuarantineBehavior == RealTime.Config.QuarantineBehavior.Family))
                {
                    CheckForContacts(citizenID, Config.DiseaseDuration, Config.QuarantineBehavior);
                    continue;
                }

                if (CitizenProxy.IsSick(ref citizens[retrieveID(citizenID)]))
                {
                    numberOfInfectiousCitizens++;

                    if (!IsInfectious(citizenID) && QUICK_SIMULATION)
                    {
                        continue;
                    }

                    if (IsInfectious(citizenID))
                    {
                        Observer.AddInfectiousCitizen(citizenID);
                    }

                    if (CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)]) == 0 && CitizenProxy.GetVehicle(ref citizens[retrieveID(citizenID)]) == 0 && instanceID > 0)
                    {
                        if (!ShouldBeInQuarantine(citizenID) || Config.QuarantineBehavior == RealTime.Config.QuarantineBehavior.None) {
                            Vector3 location = CitizenMgr.GetCitizenPosition(instanceID);

                            for (int j = 0; j < initialPopulationHealthy.Count; j++)
                            {
                                uint otherCitizenID = initialPopulationHealthy[j];
                                if (ShouldBeInQuarantine(otherCitizenID))
                                {
                                    continue;
                                }

                                ushort otherInstanceID = CitizenProxy.GetInstance(ref citizens[retrieveID(otherCitizenID)]);
                                if (CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)]) == 0 && !CitizenProxy.IsSick(ref citizens[retrieveID(otherCitizenID)]) && otherInstanceID > 0)
                                {
                                    Vector3 otherLocation = CitizenMgr.GetCitizenPosition(otherInstanceID);

                                    if ((otherLocation - location).magnitude < range && Masks.GetOutdoorInfectionProbability(citizenID, otherCitizenID) > random.NextDouble())
                                    {
                                        ContactManager.Instance.AddContact(citizenID, otherCitizenID, false, currentDateTime);

                                        if (IsInfectious(citizenID))
                                        {
                                            Debug.Log("Citizen " + otherCitizenID + "(" + otherLocation + ")" + " is in transmission range of citizen " + citizenID + "(" + TestManager.Instance.GetPositiveDate(citizenID) + ") and got infected!");

                                            Observer.AddCitizenInfection(citizenID, otherCitizenID, InfectionType.OUTDOOR, currentDateTime, otherLocation);

                                            newlyInfectedCitizenIDs.Add(otherCitizenID);
                                            InfectCitizen(otherCitizenID, ref citizens[retrieveID(otherCitizenID)]);
                                            j--;
                                            continue;
                                        }
                                    }
                                }
                            }
                        }
                    } else if (CitizenProxy.GetVehicle(ref citizens[retrieveID(citizenID)]) > 0)
                    {
                        if (!ShouldBeInQuarantine(citizenID) || Config.QuarantineBehavior == RealTime.Config.QuarantineBehavior.None)
                        {
                            for (int j = 0; j < initialPopulationHealthy.Count; j++)
                            {
                                uint otherCitizenID = initialPopulationHealthy[j];
                                if (ShouldBeInQuarantine(otherCitizenID))
                                {
                                    continue;
                                }

                                if (CitizenProxy.GetVehicle(ref citizens[retrieveID(otherCitizenID)]) > 0 && !CitizenProxy.IsSick(ref citizens[retrieveID(otherCitizenID)]))
                                {
                                    if (CitizenProxy.GetVehicle(ref citizens[retrieveID(citizenID)]) == CitizenProxy.GetVehicle(ref citizens[retrieveID(otherCitizenID)]))
                                    {
                                        ContactManager.Instance.AddContact(citizenID, otherCitizenID, false, currentDateTime);

                                        if (IsInfectious(citizenID))
                                        {
                                            if (Masks.GetVehicleInfectionProbability(citizenID, otherCitizenID) > random.NextDouble())
                                            {
                                                Debug.Log("Citizen " + otherCitizenID + " is in the same vehicle as citizen " + citizenID + "(" + TestManager.Instance.GetPositiveDate(citizenID) + ")" + " and got infected!");

                                                ushort otherInstanceID = CitizenProxy.GetInstance(ref citizens[retrieveID(otherCitizenID)]);
                                                if (otherInstanceID > 0)
                                                {
                                                    Observer.AddCitizenInfection(citizenID, otherCitizenID, InfectionType.VEHICLE, currentDateTime, CitizenMgr.GetCitizenPosition(otherInstanceID));
                                                } else
                                                {
                                                    Observer.AddCitizenInfection(citizenID, otherCitizenID, InfectionType.VEHICLE, currentDateTime, Vector3.zero);
                                                }

                                                newlyInfectedCitizenIDs.Add(otherCitizenID);
                                                InfectCitizen(otherCitizenID, ref citizens[retrieveID(otherCitizenID)]);
                                                j--;
                                                continue;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)]) > 0)
                    {
                        for (int j = 0; j < initialPopulationHealthy.Count; j++)
                        {
                            uint otherCitizenID = initialPopulationHealthy[j];
                            if (ShouldBeInQuarantine(otherCitizenID))
                            {
                                continue;
                            }

                            if (CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(otherCitizenID)]) > 0 && !CitizenProxy.IsSick(ref citizens[retrieveID(otherCitizenID)]))
                            {
                                if (CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)]) == CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(otherCitizenID)]))
                                {
                                    ContactManager.Instance.AddContact(citizenID, otherCitizenID, true, currentDateTime);

                                    if (IsInfectious(citizenID))
                                    {
                                        double probability;
                                        bool infect = false;
                                        if (CitizenProxy.GetHomeBuilding(ref citizens[retrieveID(citizenID)]) == CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)])
                                            && CitizenProxy.GetHomeBuilding(ref citizens[retrieveID(otherCitizenID)]) == CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(otherCitizenID)]))
                                        {
                                            uint[] family = new uint[10];
                                            if (CitizenMgr.TryGetFamily(citizenID, family))
                                            {
                                                if (Array.Exists(family, element => element == otherCitizenID))
                                                {
                                                    infect = !ShouldBeInQuarantine(citizenID);
                                                    probability = Masks.GetDefaultIndoorInfectionProbability();
                                                } else
                                                {
                                                    infect = !ShouldBeInQuarantine(citizenID);
                                                    probability = Masks.GetIndoorInfectionProbabilityNoContact();
                                                }
                                            } else
                                            {
                                                infect = !ShouldBeInQuarantine(citizenID);
                                                probability = Masks.GetIndoorInfectionProbabilityNoContact();
                                            }

                                        }
                                        else
                                        {
                                            infect = !ShouldBeInQuarantine(citizenID) || Config.QuarantineBehavior == RealTime.Config.QuarantineBehavior.None;
                                            probability = Masks.GetIndoorInfectionProbability(citizenID, otherCitizenID);
                                        }
                                        if (infect)
                                        {
                                            if (probability > random.NextDouble())
                                            {
                                                Debug.Log("Citizen " + otherCitizenID + " is in the same building as citizen " + citizenID + "(" + TestManager.Instance.GetPositiveDate(citizenID) + ")" + " and got infected!");

                                                ushort buildingID = CitizenProxy.GetCurrentBuilding(ref citizens[retrieveID(citizenID)]);
                                                Building building = BuildingManager.instance.m_buildings.m_buffer[buildingID];

                                                Observer.AddCitizenInfection(citizenID, otherCitizenID, InfectionType.INDOOR, currentDateTime, building.m_position, BuildingMgr.GetBuildingService(buildingID));

                                                newlyInfectedCitizenIDs.Add(otherCitizenID);
                                                InfectCitizen(otherCitizenID, ref citizens[retrieveID(otherCitizenID)]);
                                                j--;
                                                continue;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

            }
        }

        private uint retrieveID(uint citizenID)
        {
            if (citizenMatching.ContainsKey(citizenID))
            {
                return citizenMatching[citizenID];
            }
            return citizenID;
        }

        public bool ShouldBeInQuarantine(uint citizenID)
        {
            if (IsInQuarantine(citizenID))
            {
                return true;
            }

            Citizen citizen = CitizenMgr.GetCitizensArray()[citizenID];
            if (CitizenProxy.IsSick(ref citizen) && Config.QuarantineBehavior != RealTime.Config.QuarantineBehavior.None)
            {
                long infectionInDays = ((simulation.m_currentGameTime.Ticks / 10000) - GetInfectionDate(citizenID)) / 1000 / 3600 / 24;
                bool knownSick = IsKnownSick(citizenID);
                if ((knownSick && !Config.OnlyTestedCitizensToQuarantine) || TestManager.Instance.IsBlocked(citizenID, currentDateTime) || infectionInDays > Config.DetectionTime)
                {
                    if (!IsInQuarantine(citizenID))
                    {
                        QuarantineManager.Instance.AddCitizenInQuarantine(citizenID, currentDateTime);
                    }
                    return true;
                }
            }

            return false;
        }

        private bool IsKnownSick(uint citizenID)
        {
            Citizen citizen = CitizenMgr.GetCitizensArray()[citizenID];
            if (CitizenProxy.IsSick(ref citizen))
            {
                long infectionInDays = ((simulation.m_currentGameTime.Ticks / 10000) - GetInfectionDate(citizenID)) / 1000 / 3600 / 24;
                bool hasSymptoms = HasSymptoms(citizenID);
                return infectionInDays > Config.StartSymptoms && hasSymptoms;
            }
            return false;
        }

        private void CheckForContacts(uint citizenID, uint timeInDays, Config.QuarantineBehavior quarantineBehavior)
        {
            if (QuarantineManager.Instance.HasBeenChecked(citizenID))
            {
                return;
            }

            if (!(TestManager.Instance.IsTestedPositive(citizenID, currentDateTime) || (!Config.OnlyTestedCitizensToQuarantine && IsKnownSick(citizenID))))
            {
                return;
            }

            QuarantineManager.Instance.AddCheckedCitizen(citizenID);

            if (quarantineBehavior == RealTime.Config.QuarantineBehavior.Contacts)
            {
                Dictionary<uint, DateTime> contacts = ContactManager.Instance.GetContactsForCitizen(citizenID);

                if (contacts == null)
                {
                    return;
                }

                long leastConsideredTime = simulation.m_currentGameTime.Ticks - (TimeSpan.TicksPerDay * timeInDays);

                foreach (KeyValuePair<uint, DateTime> contact in contacts)
                {
                    if (contact.Value.Ticks > leastConsideredTime)
                    {
                        QuarantineManager.Instance.AddCitizenInProphylacticQuarantine(contact.Key, currentDateTime);
                    }
                }
            }

            if (quarantineBehavior == RealTime.Config.QuarantineBehavior.Family)// || quarantineBehavior == RealTime.Config.QuarantineBehavior.Contacts)
            {
                uint[] family = new uint[10];
                if (CitizenMgr.TryGetFamily(citizenID, family))
                {
                    foreach (uint otherCitizenID in family)
                    {
                        if (otherCitizenID > 0)
                        {
                            QuarantineManager.Instance.AddCitizenInProphylacticQuarantine(otherCitizenID, currentDateTime);
                        }
                    }
                }
            }
        }

        internal bool HasSymptoms(uint citizenID)
        {
            return infectedCitizensWithSymptoms.Contains(citizenID);
        }

        public void kill(double childDeathProbability, double teenDeathProbability, double youngDeathProbability, double adultDeathProbability, double seniorDeathProbability)
        {
            Citizen[] citizens = CitizenMgr.GetCitizensArray();
            
            for (int i = 0; i < initialPopulationSick.Count; i++)
            {
                uint citizenID = initialPopulationSick[i];
                if (!activeInfections.ContainsKey(citizenID))
                {
                    continue;
                }
                float infectedTime = ((currentDateTime.Ticks / 10000) - activeInfections[citizenID]) / (24 * 3600 * 1000f);
                
                if (infectedTime < Config.StartSymptoms)
                {
                    continue;
                }

                double deathProbability = 0;
                Citizen.AgeGroup ageGroup = CitizenProxy.GetAge(ref citizens[retrieveID(citizenID)]);
                switch (ageGroup)
                {
                    case Citizen.AgeGroup.Child:
                        deathProbability = childDeathProbability;
                        break;
                    case Citizen.AgeGroup.Teen:
                        deathProbability = teenDeathProbability;
                        break;
                    case Citizen.AgeGroup.Young:
                        deathProbability = youngDeathProbability;
                        break;
                    case Citizen.AgeGroup.Adult:
                        deathProbability = adultDeathProbability;
                        break;
                    case Citizen.AgeGroup.Senior:
                        deathProbability = seniorDeathProbability;
                        break;
                }

                if (random.NextDouble() < deathProbability)
                {
                    Debug.Log("Killing citizen " + citizenID);
                    KillCitizen(citizenID, ref citizens[retrieveID(citizenID)]);
                    i--;
                }
            }

        }

        /// <summary>
        /// Gets a reference to the citizen manager proxy object.
        /// </summary>
        private ICitizenManagerConnection<Citizen> CitizenMgr { get; set; }

        /// <summary>
        /// Gets a reference to the proxy class that provides access to citizen's methods and fields.
        /// </summary>
        private ICitizenConnection<Citizen> CitizenProxy { get; set; }

        /// <summary>
        /// Gets a reference to the proxy class that provides access to buildings methods and fields.
        /// </summary>
        private IBuildingManagerConnection BuildingMgr { get; set; }
        
        /// <summary>
        /// Gets or sets the configuration.
        /// </summary>
        private Config.RealTimeConfig Config { get; set; }
        public static PandemicManager Instance { get; set; }
    }
}
